(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

(* The encoding of the store; the encoding of references. *)

(* ------------------------------------------------------------------------- *)

(* A store in world [w] is an array that has shape [w x] for every future
   world [x]. That is, the store is polymorphic in its depth parameter. In
   other words, the values that are currently in the store are necessary:
   they will remain valid as we move to future worlds. *)

type store : later world -> * =
  \x. forall y. array (x y)

(* ------------------------------------------------------------------------- *)

(* A reference in world [x] is an index that can be used to read or write an
   array of type [store x]. Thus, for every future world [y], this index must
   have a type of the form [index (x y) ...]. If this is a reference of type
   [a], then the value that is read or written must have type [a (x `o` y)]. *)

type ref : stype -> stype =
  \a x.
    forall y. index (x y) (a (x `o` y))

(* ------------------------------------------------------------------------- *)

(* The above encoding of the reference type is such that the types [ref a] and
   [box (ref a)] are in fact equivalent. *)

(* The following statement spells out the definition of [box (ref a)]. *)

lemma box_ref:
  forall a x y.
  box (ref a) x =
  (forall y z. index (x (y `o` z) `@` y z) (a (x `o` y `o` z)))

(* The following coercion converts [ref a] to [box (ref a)]. The depth aspect
   of the coercion is made possible made the universal quantification over [y]
   in the definition of [ref], while the width aspect is made possible by the
   fact that a valid index into a smaller array is also a valid index into a
   larger array. *)

term box_ref: forall a x. ref a x -> box (ref a) x =
  /\ a x.
  \r : ref a x.
  /\ y z.
    index_monotonic
      [x (y `o` z)]
      [y z]
      [a (x `o` y `o` z)]
      (r [y `o` z])

(* ------------------------------------------------------------------------- *)

(* The initial store. *)

term store_empty : store nil =
  /\ x. array_empty

(* ------------------------------------------------------------------------- *)

(* Extending the store by one cell is just a matter of invoking [array_extend]
   with appropriate type parameters. Such an extension takes us from world [w]
   to world [w `o` cell a w], if the new reference that is being allocated has
   type [a]. *)

(* The following two facts may help understand the code that follows. *)

lemma store_extend_fact_1:
  forall x a.
  store (x `o` cell a x) =
  (forall y. array ((x `o` cell a x) y))

lemma store_extend_fact_2:
  forall x a.
  store (x `o` cell a x) =
  (forall y. array (x (cell a x `o` y) `snoc` a (x `o` cell a x `o` y)))

term store_extend : forall x a. store x -> box a x -> store (x `o` cell a x) =
  /\ x a.
  \s : store x.
  \v : box a x.
  /\ y.
    type cy = cell a x `o` y in
    array_extend [x cy] [a (x `o` cy)] (s [cy]) (v [cy])

(* ------------------------------------------------------------------------- *)

(* The end address of the store serves as the index of the next reference.
   In other words, creating a new reference is just a matter of invoking
   [array_end_index]. *)

(* The following fact spells out what kind of index must be produced. *)

lemma ref_compose_cell:
  forall a x.
  ref a (x `o` cell a x) =
  (forall y. index ((x `o` cell a x) y) (a (x `o` cell a x `o` y)))

term store_end_index : forall x a. store x -> ref a (x `o` cell a x) =
  /\ x a.
  \s : store x.
  /\ y.
    type cy = cell a x `o` y in
    array_end_index [x cy] (s [cy]) [a (x `o` cy)]

