Set Implicit Arguments.
Require Import MyLists.
Require Import DeBruijn.

Inductive sterm :=
| SVar: nat -> sterm
| SAbs: sterm -> sterm
| SApp: sterm -> sterm -> sterm
| SUnit
| SPair: sterm -> sterm -> sterm
| SProjLeft: sterm -> sterm
| SProjRight: sterm -> sterm
| SReturn: sterm
| SBind: sterm
| SNew: sterm
| SRead: sterm
| SWrite: sterm
| SLoc: nat -> sterm
.

Definition store :=
  list sterm.

Inductive ectx :=
| EHole: ectx
| EBind: ectx -> sterm -> ectx.

Fixpoint lift_sterm (k : nat) (t : sterm) : sterm :=
  match t with
  | SVar x =>
      SVar (lift_var k x)
  | SAbs t =>
      SAbs (lift_sterm (1 + k) t)
  | SApp t1 t2 =>
      SApp (lift_sterm k t1) (lift_sterm k t2)
  | SUnit =>
      SUnit
  | SPair t1 t2 =>
      SPair (lift_sterm k t1) (lift_sterm k t2)
  | SProjLeft t =>
      SProjLeft (lift_sterm k t)
  | SProjRight t =>
      SProjRight (lift_sterm k t)
  | SReturn
  | SBind
  | SNew
  | SRead
  | SWrite
  | SLoc _ =>
      t
  end.

Fixpoint subst_sterm (u : sterm) (k : nat) (t : sterm) : sterm :=
  match t with
  | SVar x =>
      subst_var SVar u k x
  | SAbs t =>
      SAbs (subst_sterm (lift_sterm 0 u) (1 + k) t)
  | SApp t1 t2 =>
      SApp (subst_sterm u k t1) (subst_sterm u k t2)
  | SUnit =>
      SUnit
  | SPair t1 t2 =>
      SPair (subst_sterm u k t1) (subst_sterm u k t2)
  | SProjLeft t =>
      SProjLeft (subst_sterm u k t)
  | SProjRight t =>
      SProjRight (subst_sterm u k t)
  | SReturn
  | SBind
  | SNew
  | SRead
  | SWrite
  | SLoc _ =>
      t
  end.

Lemma subst_lift_sterm:
  subst_lift lift_sterm subst_sterm.
Proof.
  subst_lift_tac.
Qed.

Local Hint Resolve subst_lift_sterm.

Lemma lift_lift_sterm:
  lift_lift lift_sterm.
Proof.
  unfold lift_lift; intro a; induction a; intros; simpl in *; f_equal; eauto with lift_lift_db.
Qed.

Local Hint Resolve lift_lift_sterm.

Lemma lift_zero_lift_sterm:
  lift_zero_lift lift_sterm.
Proof.
  lift_zero_lift_tac.
Qed.

Ltac lzl_sterm :=
  lzl lift_sterm lift_zero_lift_sterm.

Hint Extern 1 => lzl_sterm : lift_subst_1_db.
Hint Extern 1 => lzl_sterm : lift_subst_2_db.
Hint Extern 1 => lzl_sterm : subst_subst_db.

Lemma lift_subst_1_sterm:
  lift_subst_1 lift_sterm lift_sterm subst_sterm.
Proof.
  lift_subst_1_tac.
Qed.

Local Hint Resolve lift_subst_1_sterm.

Hint Extern 1 => rewrite lift_subst_1_sterm; [ idtac | omega ] : subst_subst_db.

Lemma lift_subst_2_sterm:
  lift_subst_2 lift_sterm lift_sterm subst_sterm.
Proof.
  lift_subst_2_tac.
Qed.

Local Hint Resolve lift_subst_2_sterm.

Lemma subst_subst_sterm:
  subst_subst lift_sterm subst_sterm subst_sterm.
Proof.
  subst_subst_tac.
Qed.

Local Hint Resolve subst_subst_sterm.

Lemma pun_1_sterm:
  pun_1 SVar lift_sterm subst_sterm.
Proof.
  pun_tac.
Qed.

Lemma pun_2_sterm:
  pun_2 SVar lift_sterm subst_sterm.
Proof.
  pun_tac.
Qed.

Lemma lift_sterm_injective:
  lift_injective lift_sterm.
Proof.
  lift_injective_tac.
Qed.

Definition closed_sterm :=
  closed lift_sterm.

Lemma lift_closed_sterm:
  forall k t,
  closed_sterm t ->
  lift_sterm k t = t.
Proof.
  eauto using lift_closed.
Qed.

Lemma subst_closed_sterm:
  forall u k t,
  closed_sterm t ->
  subst_sterm u k t = t.
Proof.
  eauto using subst_closed.
Qed.

Definition closed_store (s : store) :=
  forall t, In t s -> closed_sterm t.

Inductive closed_ectx : ectx -> Prop :=
| ClosedEHole:
    closed_ectx EHole
| ClosedEBind:
    forall E t,
    closed_ectx E ->
    closed_sterm t ->
    closed_ectx (EBind E t).

Hint Constructors closed_ectx.

Notation closed_conf s E t :=
  (closed_store s /\ closed_ectx E /\ closed_sterm t).

