(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

open Atom
open Terms

(* ------------------------------------------------------------------------- *)

(* This is a data type of untyped terms with variables in de Bruijn style. *)

type uterm =
  | UVar of int
  | UGlobalVar of atom
  | UAbs of uterm
  | UApp of uterm * uterm
  | ULet of uterm * uterm
  | UUnit
  | UPair of uterm * uterm
  | ULetPair of uterm * uterm

(* ------------------------------------------------------------------------- *)

(* The following functions translate FORK terms into untyped terms. Type
   annotations are dropped; atoms are converted to de Bruijn indices. *)

(* The conversion environment for local variables is a list of optional atoms;
   the options appear because FORK's [TeLetPair] construct supports
   wildcards. *)

(* Global variables remain named and are translated to Coq definitions.
   In general, FORK term definitions can be recursive: here, we assume
   that this is not the case. Global definitions represent closed terms,
   so references to them need not be lifted. *)

let rec find k x = function
  | [] ->
      UGlobalVar x
  | Some y :: env when equal x y ->
      UVar k
  | _ :: env ->
      find (k + 1) x env

let rec erase_term env = function
  | TeVar x ->
      find 0 x env
  | TeAbs (x, _, t) ->
      UAbs (erase_locterm (Some x :: env) t)
  | TeApp (t1, t2) ->
      UApp (erase_locterm env t1, erase_locterm env t2)
  | TeTyAbs (_, _, t)
  | TeTyApp (t, _)
  | TePack (t, _, _)
  | TeConstraint (t, _)
  | TeType (_, _, t) ->
      erase_locterm env t
  | TeUnpack (_, x, t1, t2)
  | TeLet (x, t1, t2) ->
      ULet (erase_locterm env t1, erase_locterm (Some x :: env) t2)
  | TeUnit ->
      UUnit
  | TePair (t1, t2) ->
      UPair (erase_locterm env t1, erase_locterm env t2)
  | TeLetPair (x, y, t1, t2) ->
      ULetPair (erase_locterm env t1, erase_locterm (y :: x :: env) t2)

and erase_locterm env (_, t) =
  erase_term env t

let erase_signature (terms : signature) : uterm AtomMap.t =
  AtomMap.map (fun (_, t) ->
    erase_locterm [] t
  ) terms

(* ------------------------------------------------------------------------- *)

(* We perform a topological sort of the toplevel term definitions, so as to
   print them in an appropriate order. If there is any cycle, we fail. *)

type color =
  | White
  | Grey
  | Black

let dfs (f : atom -> uterm -> unit) (terms: uterm AtomMap.t) : unit =

  let color =
    ref AtomMap.empty
  in

  let get_color a =
    try
      AtomMap.find a !color
    with Not_found ->
      White

  and set_color a c =
    color := AtomMap.add a c !color
  in

  let rec dfs_uterm = function
    | UVar _
    | UUnit ->
	()
    | UGlobalVar a ->
	dfs a (AtomMap.find a terms)
    | UAbs t ->
	dfs_uterm t
    | UApp (t1, t2)
    | ULet (t1, t2)
    | UPair (t1, t2)
    | ULetPair (t1, t2) ->
	dfs_uterm t1;
	dfs_uterm t2

  and dfs a t =
    match get_color a with
    | White ->
	set_color a Grey;
	dfs_uterm t;
	set_color a Black;
	f a t
    | Grey ->
	Error.signala [ a ] "Error: cannot convert to Coq: this term definition is recursive."
    | Black ->
	()
  in

  AtomMap.iter dfs terms;
  Error.signaled()

(* ------------------------------------------------------------------------- *)

(* A pretty-printer for untyped terms, in Coq syntax. *)

module P = struct

  open Printf
  open Pprint

  let (+) d1 d2 =
    d1 ^^ space ^^ d2

  (* [jump body] either displays [body] on a single line; or breaks a line,
     prints [body] at indentation 2, and breaks another line. *)

  let jump body =
    group (nest 2 (linebreak ^^ body) ^^ linebreak)

  (* Our naming convention for global definitions. *)

  let global x =
    let name = Identifier.name (Atom.identifier x) in
    match name with
    | "return" ->
	"__return" (* "return" is a Coq keyword, avoid it; yes, this is very ad hoc *)
    | _ ->
	name

  let rec uterm = function
    | UVar x ->
	text "TVar" + text (sprintf "%d" x)
    | UGlobalVar x ->
	text (global x)
    | UAbs t ->
	text "TAbs" + puterm t
    | UApp (t1, t2) ->
	text "TApp" + puterm t1 + puterm t2
    | ULet (t1, t2) ->
	text "TLet" + puterm t1 + puterm t2
    | UUnit ->
	text "TUnit"
    | UPair (t1, t2) ->
	text "TPair" + puterm t1 + puterm t2
    | ULetPair (t1, t2) ->
	text "TLetPair" + puterm t1 + puterm t2

  and juterm t =
    jump (uterm t)

  and puterm t =
    parens (juterm t)

  let juterm : uterm -> string =
    wrap juterm

  let signature (terms : uterm AtomMap.t) =
    dfs (fun x t ->
      let x = global x in
      printf "Definition %s := %s.\n\n" x (juterm t);
      printf "Lemma closed_%s:\n" x;
      printf "  closed_tterm %s.\n" x;
      printf "Proof.\n  unfold closed_tterm, closed, well_scoped_lift. auto.\nQed.\n\n";
      printf "Hint Resolve closed_%s.\n\n" x;
      printf "Lemma lift_%s:\n" x;
      printf "  forall k,\n  lift_tterm k %s = %s.\nProof.\n  auto.\nQed.\n\n" x x;
      printf "Hint Rewrite lift_%s : lift.\n\n" x;
      printf "Lemma subst_%s:\n" x;
      printf "  forall u k,\n  subst_tterm u k %s = %s.\nProof.\n  auto.\nQed.\n\n" x x;
      printf "Hint Rewrite subst_%s : subst.\n\n" x;
    ) terms

end

(* ------------------------------------------------------------------------- *)

(* Putting it all together. *)

let prologue =
  "Require Import DeBruijn.\nRequire Import TargetSyntax.\n\n(* THIS FILE IS AUTOMATICALLY GENERATED BY FORK. DO NOT EDIT! *)\n\n"

let print terms =
  print_string prologue;
  P.signature (erase_signature terms)

